import pytest

# This file can be used to define fixtures and other test configuration
# that will be available to all test files

# Enable asyncio support for pytest
pytest_plugins = ["pytest_asyncio"]
